use crate::jwt::token_factory;
use crate::state::CaptchaState;
use crate::ServerState;

use actix_web::web::Data;
use actix_web::{post, web::Json, HttpResponse, Responder, Result};
use argon2::password_hash::rand_core::{OsRng, RngCore};
use argon2::password_hash::SaltString;
use argon2::password_hash::*;
use argon2::Argon2;
use argon2::PasswordHasher;
use argon2::PasswordVerifier;
use biosvg::BiosvgBuilder;
use log::*;
use serde::{Deserialize, Serialize};

#[derive(Debug, Serialize, Deserialize)]
pub struct LoginData {
    username: String,
    password: String,
}

#[derive(Debug, Serialize, Deserialize)]
struct LoginResponse {
    username: String,
    token: String,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct RegisterData {
    username: String,
    password: String,
    captcha: String,
}

#[post("/register")]
pub async fn register(
    data: Json<RegisterData>,
    state: Data<ServerState>,
) -> Result<impl Responder> {
    // First check if the user already exists
    let exists = sqlx::query!(
        "SELECT username FROM users WHERE username = ?",
        data.username
    )
    .fetch_one(&state.pool)
    .await
    .ok()
    .map(|row| row.username);

    // Bail out if the user already exists
    if exists.is_some() {
        info!("User \"{}\" already exists", data.username);
        return Ok(HttpResponse::BadRequest().json("Error"));
    }

    // Unwrapping here because if this fails, we have a serious problem
    let phc_hash = Argon2::default()
        .hash_password(data.password.as_bytes(), &SaltString::generate(&mut OsRng))
        .unwrap()
        .to_string();

    // Insert our new user into the database
    sqlx::query!(
        "INSERT INTO users (username, password) VALUES (?, ?)",
        data.username,
        phc_hash
    )
    .execute(&state.pool)
    .await
    .unwrap();

    info!("User: {} registered", data.username);
    Ok(HttpResponse::Ok().json("User registered"))
}

#[post("/login")]
pub async fn login(data: Json<LoginData>, state: Data<ServerState>) -> Result<impl Responder> {
    let uname = data.username.clone();
    let q = sqlx::query!("SELECT password FROM users WHERE username = ?", uname)
        .fetch_one(&state.pool)
        .await
        .ok();

    if q.is_none() {
        info!("User \"{}\" failed to log in", data.username);
        return Ok(HttpResponse::BadRequest().json("Error"));
    }

    let phc_password = q.unwrap().password;
    let phc_password = PasswordHash::new(&phc_password).unwrap_or_else(|_| {
        warn!(
            "Invalid hash for user {} fetched from database (not a valid PHC string)",
            data.username
        );
        panic!();
    });

    match Argon2::default().verify_password(data.password.as_bytes(), &phc_password) {
        Ok(_) => {
            info!("User {} logged in", data.username);
            let token = token_factory(&data.username).unwrap();
            println!("{:?}", token);
            return Ok(HttpResponse::Ok().json(LoginResponse {
                username: data.username.clone(),
                token,
            }));
        }
        Err(_) => {
            info!("User \"{}\" failed to log in", data.username);
            return Ok(HttpResponse::BadRequest().json("Error"));
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
pub struct CaptchaResponse {
    captcha_svg: String,
    captcha_id: i32,
}

/// Request a captcha from the captcha service
#[post("/captcha")]
pub async fn captcha_request(cstate: Data<CaptchaState>) -> Result<impl Responder> {
    // This might block the thread a bit too long
    let (answer, svg) = get_captcha();

    let id = rand_core::OsRng.next_u32() as i32;

    let cresponse = CaptchaResponse {
        captcha_svg: svg.clone(),
        captcha_id: id,
    };

    // This is bad in about every way i can think of
    // It might just be better to hit the database every time, and let the database
    // handle rng and maybe set a trigger to delete old captchas
    match cstate.capthca_db.lock() {
        Ok(mut db) => {
            if (db.len() as i32) > 100 {
                // To prevent the database from growing too large
                // Replace with a proper LRU cache or circular buffer
                db.remove(&(id % 100)); // This is terrible
            }
            db.insert(id, answer.clone()); // We do not care about collisions
            return Ok(HttpResponse::Ok().json(cresponse));
        }
        Err(_) => {
            // This shouldnt happen
            error!("Failed to lock captcha database");
            return Ok(HttpResponse::InternalServerError().json("Error"));
        }
    }
}

/// Returns a new captcha in the form of a tuple (answer, svg)
fn get_captcha() -> (String, String) {
    BiosvgBuilder::new()
        .length(4)
        .difficulty(6)
        .colors(vec![
            "#0078D6".to_string(),
            "#aa3333".to_string(),
            "#f08012".to_string(),
            "#33aa00".to_string(),
            "#aa33aa".to_string(),
        ])
        .build()
        .unwrap()
}
